{-# LANGUAGE UndecidableInstances, MultiParamTypeClasses, FlexibleContexts, FlexibleInstances #-}

module Feldspar.Compiler.PluginArchitecture.DefaultConvert where

import Feldspar.Compiler.Imperative.Semantics
-- ===========================================================================
--  == Defaults
-- ===========================================================================

class Default t where
    defaultValue :: t
    defaultValue = error "Default value requested."

class Combine t where
    combine :: t -> t -> t
    combine = error "Default combination function used."

instance Default Int where
    defaultValue = 0
instance Combine Int where
    combine = (+)

instance Default Bool where
    defaultValue = False

instance Default () where
    defaultValue = ()
instance Combine () where
    combine _ _ = ()

instance (Default a, Default b) => Default (a,b) where
    defaultValue = (defaultValue, defaultValue)

class Convert a b where
    convert :: a -> b

{-instance Convert a a where
    convert = id-}

instance Default b => Convert a b where
    convert _ = defaultValue
    
class (SemanticInfo from, SemanticInfo to
    , Convert (ProcedureInfo from) (ProcedureInfo to)           
    , Convert (BlockInfo from) (BlockInfo to)
    , Default (ProgramInfo to)
    , Convert (EmptyInfo from) (EmptyInfo to)
    , Convert (PrimitiveInfo from) (PrimitiveInfo to)           
    , Convert (SequenceInfo from) (SequenceInfo to)
    , Convert (BranchInfo from)    (BranchInfo to)              
    , Convert (FormalParameterInfo from) (FormalParameterInfo to)
    , Convert (SequentialLoopInfo from) (SequentialLoopInfo to) 
    , Convert (ParallelLoopInfo from) (ParallelLoopInfo to)
    , Convert (LocalDeclarationInfo from) (LocalDeclarationInfo to)
    , Convert (LeftValueExpressionInfo from) (LeftValueExpressionInfo to)
    , Convert (InputActualParameterInfo from) (InputActualParameterInfo to)
    , Convert (OutputActualParameterInfo from) (OutputActualParameterInfo to)
    , Convert (VariableInLeftValueInfo from) (VariableInLeftValueInfo to)
    , Convert (ArrayElemReferenceInfo from) (ArrayElemReferenceInfo to)
    , Convert (ProcedureCallInfo from) (ProcedureCallInfo to)
    , Convert (AssignmentInfo from) (AssignmentInfo to)
    , Convert (FunctionCallInfo from) (FunctionCallInfo to)
    , Convert (IntConstantInfo from) (IntConstantInfo to)
    , Convert (FloatConstantInfo from) (FloatConstantInfo to)
    , Convert (BoolConstantInfo from) (BoolConstantInfo to)
    , Convert (ArrayConstantInfo from) (ArrayConstantInfo to)
    , Convert (VariableInfo from) (VariableInfo to)) => ConvertAllInfos from to

instance (SemanticInfo from, SemanticInfo to -- TODO general instance needs UndecidableInstances, check whether it is ok
    , Convert (ProcedureInfo from) (ProcedureInfo to)           
    , Convert (BlockInfo from) (BlockInfo to)
    , Default (ProgramInfo to)
    , Convert (EmptyInfo from) (EmptyInfo to)
    , Convert (PrimitiveInfo from) (PrimitiveInfo to)           
    , Convert (SequenceInfo from) (SequenceInfo to)
    , Convert (BranchInfo from)    (BranchInfo to)              
    , Convert (FormalParameterInfo from) (FormalParameterInfo to)
    , Convert (SequentialLoopInfo from) (SequentialLoopInfo to) 
    , Convert (ParallelLoopInfo from) (ParallelLoopInfo to)
    , Convert (LocalDeclarationInfo from) (LocalDeclarationInfo to)
    , Convert (LeftValueExpressionInfo from) (LeftValueExpressionInfo to)
    , Convert (InputActualParameterInfo from) (InputActualParameterInfo to)
    , Convert (OutputActualParameterInfo from) (OutputActualParameterInfo to)
    , Convert (VariableInLeftValueInfo from) (VariableInLeftValueInfo to)
    , Convert (ArrayElemReferenceInfo from) (ArrayElemReferenceInfo to)
    , Convert (ProcedureCallInfo from) (ProcedureCallInfo to)
    , Convert (AssignmentInfo from) (AssignmentInfo to)
    , Convert (FunctionCallInfo from) (FunctionCallInfo to)
    , Convert (IntConstantInfo from) (IntConstantInfo to)
    , Convert (FloatConstantInfo from) (FloatConstantInfo to)
    , Convert (BoolConstantInfo from) (BoolConstantInfo to)
    , Convert (ArrayConstantInfo from) (ArrayConstantInfo to)
    , Convert (VariableInfo from) (VariableInfo to)) => ConvertAllInfos from to
